/*======================================================================================
    Space Y: An Asteroids Clone

    Copyright 2022 by TheDummyProgrammer (https://www.thedummyprogrammer.com)

    This file is part of Space Y.

    Space Y is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Space Y is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Space Y.  If not, see <http://www.gnu.org/licenses/>.
======================================================================================*/


class Point {
    private _x: number;
    public get x(): number {
        return this._x;
    }

    public set x(value: number) {
        this._x = value;
    }

    private _y: number;
    public get y(): number {
        return this._y;
    }

    public set y(value: number) {
        this._y = value;
    }

    constructor(x: number, y: number) {
        this._x = x;
        this._y = y;
    }

    public toString(): string {
        return `Point: x:${this._x}, y:${this._y}`;
    }
}

class Size {
    private _width: number = 0;
    public get width(): number {
        return this._width;
    }
    public set width(value: number) {
        this._width = value;
    }

    private _height: number = 0;
    public get height(): number {
        return this._height;
    }
    public set height(value: number) {
        this._height = value;
    }
    
    constructor (width: number, height: number) {
        this._width = width;
        this._height = height;
    }

    public toString(): string {
        return `Size: width:${this._width}, height:${this._height}`;
    }
}

class Rectangle {
    private _x: number = 0;
    public get x(): number {
        return this._x;
    }
    public set x(value: number) {
        this._x = value;
    }

    private _y: number = 0;
    public get y(): number {
        return this._y;
    }
    public set y(value: number) {
        this._y = value;
    }

    private _width: number = 0;
    public get width(): number {
        return this._width;
    }
    public set width(value: number) {
        this._width = value;
    }

    private _height: number = 0;
    public get height(): number {
        return this._height;
    }
    public set height(value: number) {
        this._height = value;
    }

    public get topLeft(): Point {
        return new Point(this.x, this.y);
    }

    public get topRight(): Point {
        return new Point(this.x + this.width, this.y);
    }

    public get bottomRight(): Point {
        return new Point(this.x + this.width, this.y + this.height);
    }

    public get bottomLeft(): Point {
        return new Point(this.x, this.y + this.height);
    }

    public offset(x: number, y: number): void {
        this.x += x;
        this.y += y;
    }

    public clone(): Rectangle {
        return new Rectangle(new Point(this._x, this._y), new Size(this._width, this._height));
    }

    constructor(location: Point, size: Size) {
        this._x = location.x;
        this._y = location.y;
        this._width = size.width;
        this._height = size.height;
    }

    public contains(pt: Point): boolean {
        return (pt.x >= this._x && pt.x <= (this._x + this._width) && pt.y >= this._y && pt.y <= (this._y + this._height));
    }

    public intersectsWith(rect: Rectangle): boolean {
        return (this.contains(rect.topLeft) || this.contains(rect.topRight) || this.contains(rect.bottomRight) || this.contains(rect.bottomLeft))
    }

    public toString(): string {
        return `Rectangle: x:${this._x}, y:${this._y}, width:${this._width}, height:${this._height}`;
    }
}

class AppParameters {
    public static gameAreaUpdateDelayMS: number = 20;
    public static accelerationIncrement: number = 0.3;
    public static accelerationDecrement: number = 0.1;
    public static rotationIncrement: number = 4;
    public static rotationDecrement: number = 4;
    public static maxSpeed: number = 12;
    public static bulletTimeToLive: number = 60;
    public static bulletInitialSpeed: number = 14;
    public static bulletInitialRotation: number = 90;
    public static boundingBoxDecrementPercent: number = 20;
    public static shipInitialRotation: number = 90;
    public static asteroidRotationIncrement: number = 1;
    public static maxUfoShips: number = 3;
    public static maxAsteroids: number = 5;
    public static maxPlayerLives: number = 3;

    public static scoreAsteroidBig: number = 20;
    public static scoreAsteroidMedium: number = 50;
    public static scoreAsteroidSmall: number = 100;
    public static scoreUfoShip: number = 200;

    public static debugDrawBoundingBoxLocal: boolean = false;
    public static debugDrawBoundingBox: boolean = false;
}

class AxisTranslation {
    private _origin: Point;
    public get origin(): Point {
        return this._origin;
    }

    public set origin(value: Point) {
        this._origin = value;
    }

    constructor(origin: Point) {
        this._origin = origin;
    }

    public toExtPoint(pt: Point): Point {
        return new Point(this._origin.x + pt.x, this._origin.y - pt.y);
    }

    public toIntPoint(pt: Point): Point {
        return new Point(pt.x - this._origin.x, this._origin.y - pt.y);
    }
}

class Utility {
    public static degreeToRadians(degree: number): number {
        return degree * (Math.PI / 180);
    }

    public static radiansToDegree(radians: number): number {
        return radians * (180 / Math.PI);
    }

    public static transformPoint(point: Point, rotation: number, zoom: number): Point {
        let radRotation: number = Utility.degreeToRadians(rotation - 90);

        let newX: number = (point.x * Math.cos(radRotation) - point.y * Math.sin(radRotation)) * zoom;
        let newY: number = (point.x * Math.sin(radRotation) + point.y * Math.cos(radRotation)) * zoom;
        
        return new Point(newX, newY);
    }

    public static getRandomInteger(min: number, max: number) {
        return Math.floor(min + Math.random() * (max - min + 1));
    }
}

class DrawHelper {
    private _ctx: CanvasRenderingContext2D;
    public get drawContext(): CanvasRenderingContext2D {
        return this._ctx;
    }

    public set drawContext(value: CanvasRenderingContext2D) {
        this._ctx = value;
    }

    private _axisTrn: AxisTranslation;
    public get axisTranslation(): AxisTranslation {
        return this._axisTrn;
    }

    public set axisTranslation(value: AxisTranslation) {
        this._axisTrn = value;
    }

    constructor(canvasContext: CanvasRenderingContext2D, origin: Point) {
        this._ctx = canvasContext;
        this._axisTrn = new AxisTranslation(origin);
    }

    public beginPath(): void {
        this._ctx.beginPath();
    }

    public moveTo(pt: Point): void {
        let extPt: Point = this._axisTrn.toExtPoint(pt);
        this._ctx.moveTo(extPt.x, extPt.y);
    }

    public lineTo(pt: Point): void {
        let extPt: Point = this._axisTrn.toExtPoint(pt);
        this._ctx.lineTo(extPt.x, extPt.y);
    }

    public rect(rect: Rectangle): void {
        let topLeft: Point = this._axisTrn.toExtPoint(rect.topLeft);
        this._ctx.rect(topLeft.x, topLeft.y, rect.width, rect.height);
    }

    public get strokeStyle(): string | CanvasGradient | CanvasPattern {
        return this._ctx.strokeStyle;
    }

    public set strokeStyle(value: string | CanvasGradient | CanvasPattern) {
        this._ctx.strokeStyle = value;
    }

    public stroke(): void {
        this._ctx.stroke();
    }
}

class Vector {
    constructor(magnitude: number, direction: number) {
        this._magnitude = magnitude;
        this._direction = direction;
    }

    protected _magnitude: number = 0;
    public get magnitude(): number {
        return this._magnitude;
    }

    public set magnitude(value: number) {
        if (value > 0)
            this._magnitude = value;
        else
            this._magnitude = 0;
    }

    protected _direction: number = 0;
    public get direction(): number {
        return this._direction;
    }

    public get directionRad(): number {
        return Utility.degreeToRadians(this._direction);
    }

    public set direction(value: number) {
        if (value >= 0 && value <= 360)
            this._direction = value;
        else if (value > 360) 
            this._direction = value % 360;
        else if (value < 0)
            this._direction = (360 - (value % 360));
    }

    public toPoint(): Point {
        return new Point(this._magnitude * Math.cos(this.directionRad), this._magnitude * Math.sin(this.directionRad));
    }

    public add(vectorToAdd: Vector): void {
        let thisVectorPoint: Point = this.toPoint();
        let otherVectorPoint: Point = vectorToAdd.toPoint();

        let newX: number = thisVectorPoint.x + otherVectorPoint.x;
        let newY: number = thisVectorPoint.y + otherVectorPoint.y;

        let newMagnitude: number = Math.sqrt((newX * newX) + (newY* newY));
        let newDirection: number = 0;

        if (newX === 0 && newY > 0) {
            newDirection = 90;
        }
        else if (newX === 0 && newY < 0) {
            newDirection = 270;
        }
        else if (newX === 0 && newY === 0) {
            newDirection = 0;
        }
        else if (newX > 0 && newY >= 0) {
            newDirection = Math.atan(newY / newX);
        }
        else if (newX > 0 && newY < 0) {
            newDirection = Math.atan(newY / newX) + (Math.PI * 2);
        }
        else if (newX < 0) {
            newDirection = Math.atan(newY / newX) + Math.PI;
        }

        this._magnitude = newMagnitude;
        this._direction = Utility.radiansToDegree(newDirection);
    }

    public toString(): string {
        return `Vector (Magnitude: ${this._magnitude}, Direction: ${this._direction})`;
    }
}

class GameObject {
    protected _mainDrawHelper: DrawHelper;
    protected _localDrawHelper: DrawHelper;
    
    protected _position: Point = new Point(0, 0);
    public get position(): Point {
        return this._position;
    }

    public set position(value: Point) {
        this._position = value;
    }

    protected _rotation: number = AppParameters.shipInitialRotation;
    public get rotation(): number {
        return this._rotation;
    }

    public set rotation(value: number) {
        if (value >= 0 && value <= 360)
            this._rotation = value;
        else if (value > 360) 
            this._rotation = value % 360;
        else if (value < 0)
            this._rotation = (360 - (value % 360));
    }
    
    
    protected _vectorMotionChange: Vector = null;
    public get vectorMotionChange(): Vector {
        return this._vectorMotionChange;
    }

    public set vectorMotionChange(value: Vector) {
        this._vectorMotionChange = value;
    }

    protected _vectorMotion: Vector = null;
    public get vectorMotion(): Vector {
        return this._vectorMotion;
    }

    public set vectorMotion(value: Vector) {
        this._vectorMotion = value;
    }

    protected _zoom: number = 1;  
    public get zoom(): number {
        return this._zoom;
    }

    public set zoom(value: number) {
        this._zoom = value;
    }

    protected _destroyed: boolean = false;
    public get destroyed(): boolean {
        return this._destroyed;
    }

    public set destroyed(value: boolean) {
        this._destroyed = value;
    }

    protected _lines: Set<Point[]> = new Set<Point[]>();
    protected _polygons: Set<Point[]> = new Set<Point[]>();
    
    constructor(mainDrawHelper: DrawHelper,  position: Point) {
        this._mainDrawHelper = mainDrawHelper; 
        this._position = position; 
    }

    protected buildObject(): void {

    }

    public init(): void {
        this.buildObject();
    }

    public updatePosition(): void {
        
    }

    public getLocalBoundingBox(): Rectangle {
        let upLeftPoint: Point = null;
        let bottomRightPoint: Point = null;

        this._lines.forEach((value: Point[], key: Point[], s: Set<Point[]>) => {
            let pointsTransformed: Point[] = value.map((v: Point, i: number, a: Point[]): Point => {
                return Utility.transformPoint(v, this._rotation, this._zoom);
            });

            pointsTransformed.forEach((v: Point, i: number, a: Point[]): void => {
                if (upLeftPoint === null) {
                    upLeftPoint = new Point(v.x, v.y);
                    bottomRightPoint = new Point(v.x, v.y);
                }
                else {
                    if (v.x < upLeftPoint.x) upLeftPoint.x = v.x;
                    if (v.y > upLeftPoint.y) upLeftPoint.y = v.y;
                    if (v.x > bottomRightPoint.x) bottomRightPoint.x = v.x;
                    if (v.y < bottomRightPoint.y) bottomRightPoint.y = v.y;
                }
            });
        });

        this._polygons.forEach((value: Point[], key: Point[], s: Set<Point[]>) => {
            let pointsTransformed: Point[] = value.map((v: Point, i: number, a: Point[]): Point => {
                return Utility.transformPoint(v, this._rotation, this._zoom);
            });

            pointsTransformed.forEach((v: Point, i: number, a: Point[]): void => {
                if (upLeftPoint === null) {
                    upLeftPoint = new Point(v.x, v.y);
                    bottomRightPoint = new Point(v.x, v.y);
                }
                else {
                    if (v.x < upLeftPoint.x) upLeftPoint.x = v.x;
                    if (v.y > upLeftPoint.y) upLeftPoint.y = v.y;
                    if (v.x > bottomRightPoint.x) bottomRightPoint.x = v.x;
                    if (v.y < bottomRightPoint.y) bottomRightPoint.y = v.y;
                }
            });
        });

        // Reduce the bounding box
        upLeftPoint.x *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        upLeftPoint.y *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        bottomRightPoint.x *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        bottomRightPoint.y *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        
        return new Rectangle(upLeftPoint, new Size(Math.abs(bottomRightPoint.x - upLeftPoint.x), Math.abs(bottomRightPoint.y - upLeftPoint.y)));
    }

    public getBoundingBox(): Rectangle {
        let localBBox: Rectangle = this.getLocalBoundingBox();

        let extPoint: Point = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        let localAxisTranslation: AxisTranslation = new AxisTranslation(extPoint);
        
        return new Rectangle(localAxisTranslation.toExtPoint(localBBox.topLeft), new Size(localBBox.width, localBBox.height));
    }

    private drawLine(points: Point[]) {
        let pointsTransformed: Point[] = points.map((v: Point, i: number, a: Point[]): Point => {
            return Utility.transformPoint(v, this._rotation, this._zoom);
        });

        let extPoint: Point = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        this._localDrawHelper = new DrawHelper(this._mainDrawHelper.drawContext, extPoint);

        this._localDrawHelper.beginPath();

        for (let idxPointFrom: number = 0; idxPointFrom < pointsTransformed.length; idxPointFrom++) {
            let idxPointTo: number = idxPointFrom + 1;
            if (idxPointFrom === (points.length - 1))
                break;
            
            this._localDrawHelper.moveTo(pointsTransformed[idxPointFrom]);
            this._localDrawHelper.lineTo(pointsTransformed[idxPointTo]);
        }
        this._localDrawHelper.strokeStyle = 'white';
        this._localDrawHelper.stroke();
    }

    private drawPolygon(points: Point[]) {
        let pointsTransformed: Point[] = points.map((v: Point, i: number, a: Point[]): Point => {
            return Utility.transformPoint(v, this._rotation, this._zoom);
        });

        let extPoint: Point = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        this._localDrawHelper = new DrawHelper(this._mainDrawHelper.drawContext, extPoint);

        this._localDrawHelper.beginPath();

        for (let idxPointFrom: number = 0; idxPointFrom < pointsTransformed.length; idxPointFrom++) {
            let idxPointTo: number = idxPointFrom + 1;
            if (idxPointFrom === (points.length - 1))
                idxPointTo = 0;
            
            this._localDrawHelper.moveTo(pointsTransformed[idxPointFrom]);
            this._localDrawHelper.lineTo(pointsTransformed[idxPointTo]);
        }
        this._localDrawHelper.strokeStyle = 'white';
        this._localDrawHelper.stroke();
    }

    public draw(): void {
        this._lines.forEach((value: Point[], key: Point[], s: Set<Point[]>) => {
            this.drawLine(value);
        });

        this._polygons.forEach((value: Point[], key: Point[], s: Set<Point[]>) => {
            this.drawPolygon(value);
        });

        if (AppParameters.debugDrawBoundingBoxLocal) {
            let localBBox: Rectangle = this.getLocalBoundingBox();

            this._localDrawHelper.beginPath();
            this._localDrawHelper.strokeStyle = 'red';
            this._localDrawHelper.rect(localBBox);
            this._localDrawHelper.stroke();
        }
    }
}


class Asteroid extends GameObject {
    private _rotationIncrement: number = AppParameters.asteroidRotationIncrement;

    private _type: number = 0;
    public get type(): number {
        return this._type;
    }

    public set type(value: number) {
        this._type = value;
    }

    private _size: number = 1;
    public get size(): number {
        return this._size;
    }

    public set size(value: number) {
        this._zoom = value;
        this._size = value;
    }

    constructor(mainDrawHelper: DrawHelper,  position: Point) {
        super(mainDrawHelper, position);
        this._vectorMotion = new Vector(Utility.getRandomInteger(1, 3), Utility.getRandomInteger(0, 360));
        this._rotation = Utility.getRandomInteger(0, 360);
        this._rotationIncrement = Utility.getRandomInteger(0, 1);
        this._size = Utility.getRandomInteger(1, 3);
        this._zoom = this._size;
        this._type = Utility.getRandomInteger(0, 4);
    }

    protected buildObject(): void {
        let points: Point[];
        if (this._type === 0) {
            points = [new Point(-2, 12), new Point(6, 12), new Point(10, 6), new Point(10, 2), new Point(4, -10), 
                            new Point(-4, -10), new Point(0, 0), new Point(-8, -10), new Point(-12, -4), new Point(-6, -2),
                            new Point(-10, 2)];
        }
        else if (this._type === 1) {
            points = [new Point(-4, 10), new Point(4, 10), new Point(10, 2), new Point(6, -4), new Point(8, -10), 
                            new Point(-2, -10), new Point(-8, -4), new Point(-8, 6)];
        }
        else if (this._type === 2) {
            points = [new Point(-4, 10), new Point(10, 6), new Point(14, 6), new Point(12, -8), new Point(-2, -12), 
                            new Point(-10, -4), new Point(-4, 2)];
        }
        else if (this._type === 3) {
            points = [new Point(6, 10), new Point(-6, 10), new Point(-10, 6), new Point(-10, 0), new Point(-10, -6), 
                            new Point(-6, -10), new Point(6, -10), new Point(6, 0), new Point(-2, 0), new Point(-2, -4),
                            new Point(2, -4), new Point(2, -6), new Point(-4, -6), new Point(-6, -4), new Point(-6, 4),
                            new Point(-4, 6), new Point(6, 6)];
        }
        else {
            points = [new Point(-4, 8), new Point(8, 8), new Point(8, 4), new Point(0, 4), new Point(0, -0), 
                            new Point(6, 0), new Point(6, -2), new Point(0, -2), new Point(0, -10), new Point(-4, -10)];
        }

        this._polygons.add(points);
    }

    public updatePosition(): void {
        let deltaX: number = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY: number = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);

        this._position.x += deltaX;
        this._position.y += deltaY;

        if (this._rotationIncrement === 1)
            this._rotation++;
        else
            this._rotation--;
    }
}

class Bullet extends GameObject {
    private _timeToLive: number = AppParameters.bulletTimeToLive;
    constructor(mainDrawHelper: DrawHelper,  position: Point) {
        super(mainDrawHelper, position);
        this._zoom = 1;
        this._vectorMotion = new Vector(AppParameters.bulletInitialSpeed, AppParameters.bulletInitialRotation);
        this._rotation = AppParameters.bulletInitialRotation;
    }

    protected buildObject(): void {
        let points: Point[] = [new Point(0, 0), new Point(0, 3)];
        this._lines.add(points);
    }

    public updatePosition(): void {
        let deltaX: number = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY: number = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);

        this._position.x += deltaX;
        this._position.y += deltaY;

        this._timeToLive--;
        if (this._timeToLive <= 0)
            this._destroyed = true;
    }
}

class UfoBullet extends Bullet {

}

class UfoShip extends GameObject {
    private _gunPosition: Point;

    constructor(mainDrawHelper: DrawHelper,  position: Point) {
        super(mainDrawHelper, position);
        this._vectorMotion = new Vector(Utility.getRandomInteger(1, 3), Utility.getRandomInteger(0, 360));
        this._rotation = 90;
        this._zoom = 2;
    }

    protected buildObject(): void {
        let points: Point[] = [new Point(3, 6), new Point(6, 3), new Point(12, 0), new Point(6, -3), new Point(-6, -3), 
                        new Point(-12, 0), new Point(-6, 3), new Point(-3, 6)];
        this._polygons.add(points);

        let line1: Point[] = [new Point(-12, 0), new Point(12, 0)];
        this._lines.add(line1);

        let line2: Point[] = [new Point(-6, 3), new Point(6, 3)];
        this._lines.add(line2);

        this._gunPosition = new Point(0, -3);
    }

    public updatePosition(): void {
        let deltaX: number = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY: number = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);

        this._position.x += deltaX;
        this._position.y += deltaY;
    }

    public fire(userShipPosition: Point): Bullet {
        let bulletPos: Point = Utility.transformPoint(this._gunPosition, this._rotation, this._zoom);
        bulletPos.x = bulletPos.x + this._position.x;
        bulletPos.y = bulletPos.y + this._position.y;

        let bullet: Bullet = new UfoBullet(this._mainDrawHelper, bulletPos); 
        
        let dy: number = (userShipPosition.y - this._position.y);
        let dx: number = (userShipPosition.x - this._position.x);
        let fireAngleDegree: number = Utility.radiansToDegree(Math.atan(Math.abs(dy) / Math.abs(dx)));

        if (dx < 0 && dy > 0) 
            fireAngleDegree = 180 - fireAngleDegree;
        else if (dx < 0 && dy < 0)
            fireAngleDegree += 180;
        else if (dx > 0 && dy < 0)
            fireAngleDegree = 360 - fireAngleDegree;
        
        bullet.vectorMotion.direction = fireAngleDegree;
        bullet.rotation = bullet.vectorMotion.direction;

        if (this._vectorMotion !== null) {
            if (bullet.vectorMotion.magnitude < this._vectorMotion.magnitude)
                bullet.vectorMotion.magnitude = this._vectorMotion.magnitude;
        }
        
        bullet.init();
        
        return bullet;
    }
}

class Spaceship extends GameObject {
    private _gunPosition: Point;
    
    constructor(mainDrawHelper: DrawHelper,  position: Point) {
        super(mainDrawHelper, position);
        this._zoom = 1.5;
    }

    protected buildObject(): void {
        let points: Point[] = [new Point(-6, -4), new Point(0, 10), new Point(6, -4), new Point(5, -2), new Point(-5, -2)];
        this._polygons.add(points);
        this._gunPosition = new Point(0, 10);
    }

    public accelerate(): void {
        this._vectorMotionChange = new Vector(AppParameters.accelerationIncrement, this._rotation);
    }

    public stopAccelerate(): void {
        this._vectorMotionChange = null;
    }

    public updatePosition(): void {
        if (this._vectorMotionChange === null) {
            if (this._vectorMotion !== null)
                this._vectorMotion.magnitude -= AppParameters.accelerationDecrement;
        }
        else {
            if (this._vectorMotion === null)
                this._vectorMotion = this._vectorMotionChange;
            else
                this._vectorMotion.add(this._vectorMotionChange);
        }

        if (this._vectorMotion !== null) {
            if (this._vectorMotion.magnitude === 0)
                this._vectorMotion = null;    
            else if (this._vectorMotion.magnitude > AppParameters.maxSpeed)
                this._vectorMotion.magnitude = AppParameters.maxSpeed;
        }
        
        if (this._vectorMotion === null)
            return;

        let deltaX: number = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY: number = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);

        this._position.x += deltaX;
        this._position.y += deltaY;
    }

    public fire(): Bullet {
        let bulletPos: Point = Utility.transformPoint(this._gunPosition, this._rotation, this._zoom);
        bulletPos.x = bulletPos.x + this._position.x;
        bulletPos.y = bulletPos.y + this._position.y;

        let bullet: Bullet = new Bullet(this._mainDrawHelper, bulletPos); 
        bullet.vectorMotion.direction = this._rotation;
        bullet.rotation = this._rotation;

        if (this._vectorMotion !== null) {
            if (bullet.vectorMotion.magnitude < this._vectorMotion.magnitude)
                bullet.vectorMotion.magnitude = this._vectorMotion.magnitude;
        }
        
        bullet.init();
        
        return bullet;
    }
}

enum GameObjectType {
    playerSpaceship,
    bullet,
    ufoShip,
    ufoBullet,
    asteroid
}

enum GameMode {
    startScreen,
    play,
    gameOverScreen
}

class Game {
    private _axisOrigin: Point;
    private _body: HTMLBodyElement;
    private _canvas: HTMLCanvasElement;
    private _context: CanvasRenderingContext2D;
    private _drawHelper: DrawHelper;
    private _gameAreaWidth: number;
    private _gameAreaHeight: number;
    private _gameMode: GameMode = GameMode.startScreen;
    private _gameObjects: GameObject[] = [];
    private _keyDown: Map<string, boolean> = new Map<string, boolean>();
    private _keyUp: Map<string, boolean> = new Map<string, boolean>();
    private _spaceShip: Spaceship;
    private _fireCount: number = 0;
    private _updateGameAreaCount: number = 0;
    private _ufoShips: UfoShip[] = [];
    private _playerLives: number = AppParameters.maxPlayerLives;
    private _playerScore: number = 0;

    constructor() {
        
    }

    private _clearGameArea(): void {
        this._context.beginPath();
        this._context.fillStyle = 'black';
        this._context.fillRect(0, 0, this._canvas.width, this._canvas.height);
    }

    private _drawPlayerLives(): void {
        for (let i: number = 0; i < this._playerLives; i++) {
            let ship: Spaceship = new Spaceship(this._drawHelper, new Point(20 + i * 30, this._gameAreaHeight - 30));
            ship.zoom = 2;
            ship.init();
            ship.draw();
        }
    }

    private _drawPlayerScore(): void {
        this._context.font = '24px Courier New';
        this._context.fillStyle = 'white';
        this._context.textAlign = "left";
        this._context.fillText(this._playerScore.toString(), 20, 75);
    }

    private _manageUfoShips(): void {
        if (this._updateGameAreaCount % 1000 === 0) {
            // Add an Ufo ship when needed
            let ufoShip: UfoShip = new UfoShip(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight)));
            ufoShip.init();

            if (this._ufoShips.length < AppParameters.maxUfoShips) {
                this._gameObjects.push(ufoShip);
                this._ufoShips.push(ufoShip);
            }
        }

        for (let ufoShip of this._ufoShips) {
            if (this._updateGameAreaCount % 100 === 0) {
                let bullet: Bullet = ufoShip.fire(this._spaceShip.position);
                this._gameObjects.push(bullet);
            }
        }
    }

    private _removeUfoShip(ufoShip: UfoShip): void {
        let index = this._ufoShips.indexOf(ufoShip);

        if (index >= 0)
            this._ufoShips.splice(index, 1);
    }

    private _splitOrDestroyAsteroid(asteroid: Asteroid): void {
        if (asteroid.size > 1) {
            // Size > 1, decrease the size of the astoroid and create a new one of the same size
            asteroid.size--;
            let newAsteroid = new Asteroid(this._drawHelper, new Point(asteroid.position.x, asteroid.position.y));
            newAsteroid.size = asteroid.size;
            newAsteroid.type = asteroid.type;
            newAsteroid.rotation = asteroid.rotation;
            newAsteroid.vectorMotion.magnitude = asteroid.vectorMotion.magnitude;
            newAsteroid.vectorMotion.direction = asteroid.vectorMotion.direction - Utility.getRandomInteger(45, 270);
            newAsteroid.init();
            this._gameObjects.push(newAsteroid);
        }
        else {
            // Size = 1, destroy the asteroid
            asteroid.destroyed = true;
        }
    }

    private _getObjectType(gameObject: GameObject): GameObjectType {
        if (gameObject instanceof Spaceship)
            return GameObjectType.playerSpaceship;
        else if (gameObject instanceof UfoBullet)
            return GameObjectType.ufoBullet;
        else if (gameObject instanceof Bullet)
            return GameObjectType.bullet;
        else if (gameObject instanceof UfoShip)
            return GameObjectType.ufoShip;
        else 
            return GameObjectType.asteroid;
    }
    
    private _checkCollisions(gameObject: GameObject) {
        for (let collisionObjIdx: number = 0; collisionObjIdx < this._gameObjects.length; collisionObjIdx++) {
            let collisionObject: GameObject = this._gameObjects[collisionObjIdx];

            if (gameObject === collisionObject)
                continue;
            
            if (gameObject.destroyed)
                continue;
            
            let gameObjectType: GameObjectType = this._getObjectType(gameObject);
            let collisionObjectType: GameObjectType = this._getObjectType(collisionObject);

            if ( !((gameObjectType === GameObjectType.bullet && collisionObjectType === GameObjectType.ufoShip)
                    || (gameObjectType === GameObjectType.bullet && collisionObjectType === GameObjectType.asteroid)
                    || (gameObjectType === GameObjectType.ufoBullet && collisionObjectType === GameObjectType.playerSpaceship)
                    || (gameObjectType === GameObjectType.ufoBullet && collisionObjectType === GameObjectType.asteroid)
                    || (gameObjectType === GameObjectType.playerSpaceship && collisionObjectType === GameObjectType.ufoShip)
                    || (gameObjectType === GameObjectType.playerSpaceship && collisionObjectType === GameObjectType.asteroid)
                    || (gameObjectType === GameObjectType.ufoShip && collisionObjectType === GameObjectType.asteroid)) )
                continue;

            let bBox1: Rectangle = gameObject.getBoundingBox();
            let bBox2: Rectangle = collisionObject.getBoundingBox();

            if (AppParameters.debugDrawBoundingBox) {
                this._context.beginPath();
                this._context.strokeStyle = 'yellow';
                this._context.rect(bBox1.x, bBox1.y, bBox1.width, bBox1.height);
                this._context.rect(bBox2.x, bBox2.y, bBox2.width, bBox2.height);
                this._context.stroke();
            }

            if ( !(bBox1.intersectsWith(bBox2) || bBox2.intersectsWith(bBox1)) )
                continue;

            // The two objects intersects...
            switch (gameObjectType) {
                case GameObjectType.bullet:
                    if (collisionObjectType === GameObjectType.ufoShip) {
                        this._playerScore += AppParameters.scoreUfoShip;
                        collisionObject.destroyed = true;
                    }
                    else if (collisionObjectType === GameObjectType.asteroid) {
                        let asteroid: Asteroid = <Asteroid>collisionObject;
                        switch (asteroid.size) {
                            case 1:
                                this._playerScore += AppParameters.scoreAsteroidSmall;
                                break;
                            case 2:
                                this._playerScore += AppParameters.scoreAsteroidMedium;
                                break;
                            case 3:
                                this._playerScore += AppParameters.scoreAsteroidBig;
                                break;
                        }
    
                        this._splitOrDestroyAsteroid(asteroid);
                    }
                    break;
                
                case GameObjectType.ufoBullet:
                    if (collisionObjectType === GameObjectType.playerSpaceship) 
                        collisionObject.destroyed = true;
                    else if (collisionObjectType === GameObjectType.asteroid)
                        this._splitOrDestroyAsteroid(<Asteroid>collisionObject);
                    break;
                
                case GameObjectType.playerSpaceship:
                    if (collisionObjectType === GameObjectType.ufoShip) {
                        this._playerScore += AppParameters.scoreUfoShip;
                        collisionObject.destroyed = true;
                    }
                    else if (collisionObjectType === GameObjectType.asteroid) {
                        let asteroid: Asteroid = <Asteroid>collisionObject;
                        switch (asteroid.size) {
                            case 1:
                                this._playerScore += AppParameters.scoreAsteroidSmall;
                                break;
                            case 2:
                                this._playerScore += AppParameters.scoreAsteroidMedium;
                                break;
                            case 3:
                                this._playerScore += AppParameters.scoreAsteroidBig;
                                break;
                        }
    
                        this._splitOrDestroyAsteroid(asteroid);
                    }
                    break;
                
                case GameObjectType.ufoShip:
                    this._splitOrDestroyAsteroid(<Asteroid>collisionObject);
                    break;
            }

            gameObject.destroyed = true;

            if (collisionObjectType === GameObjectType.playerSpaceship)
                this._playerSpaceshipHit(collisionObjIdx);
            if (collisionObject.destroyed) {
                this._gameObjects.splice(collisionObjIdx, 1);
                if (collisionObject instanceof UfoShip)
                    this._removeUfoShip(collisionObject);
            }
                
        }                 
    }

    private _checkKeyDown(key: string): boolean {
        if (this._keyDown.has(key))
            return this._keyDown.get(key);
        
        return false;
    }

    private _resetKeyDown(key: string): void {
        this._keyDown.set(key, false);
    }

    private _checkKeyUp(key: string): boolean {
        if (this._keyUp.has(key)) {
            if (this._keyUp.get(key)) {
                this._keyUp.set(key, false);
                return true;    
            }
        }
        
        return false;
    }

    private _manageUserInput(): void {
        if (this._checkKeyDown("ArrowUp")) {
            this._spaceShip.accelerate();
        }
        else {
            this._spaceShip.stopAccelerate();
        }

        if (this._checkKeyDown("ArrowLeft")) {
            this._spaceShip.rotation += AppParameters.rotationDecrement;
        }

        if (this._checkKeyDown("ArrowRight")) {
            this._spaceShip.rotation -= AppParameters.rotationIncrement;
        }

        
        if (this._checkKeyDown(" ")) {
            this._fireCount++;
            if (this._fireCount === 1 || this._fireCount % 10 === 0) {
                let bullet: Bullet = this._spaceShip.fire();
                this._gameObjects.push(bullet);
            }
        }

        if (this._checkKeyUp(" ")) {
            this._fireCount = 0;
        }
    }

    private _gameModeStartScreen(): void {
        this._clearGameArea();

        if (this._checkKeyDown("Enter")) {
            this._switchToPlayMode();
            return;
        }

        for (let currentObjIdx: number = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject: GameObject = this._gameObjects[currentObjIdx];
            
            gameObject.updatePosition();
            
            gameObject.draw();
            
            if (gameObject.position.x > this._gameAreaWidth)
                gameObject.position.x = 0;
            else if (gameObject.position.x < 0)
                gameObject.position.x = this._gameAreaWidth;
            
            if (gameObject.position.y > this._gameAreaHeight)
                gameObject.position.y = 0;
            else if (gameObject.position.y < 0)
                gameObject.position.y = this._gameAreaHeight;
        }

        this._context.font = "100px Courier New";
        this._context.fillStyle = "white";
        this._context.textAlign = "center";
        this._context.fillText("Space Y", this._canvas.width / 2, this._canvas.height / 2);
        this._context.font = "25px Courier New";
        this._context.fillText("Press Enter to start", this._canvas.width / 2, (this._canvas.height / 2) + 60);
    }

    private _playerSpaceshipHit(objectIndex: number): void {
        this._spaceShip.destroyed = false;
        if (this._playerLives > 1) {
            this._playerLives--;
            this._spaceShip.position = new Point(this._gameAreaWidth / 2, this._gameAreaHeight / 2);
            this._spaceShip.rotation = 90;
            this._spaceShip.vectorMotion = null;   
            return; 
        } else {
            this._gameObjects.splice(objectIndex, 1);
            this._switchToGameOverScreenMode();
            return;
        }
    }

    private _gameModePlay(): void {
        this._updateGameAreaCount++;

        this._manageUserInput();

        this._clearGameArea();

        this._drawPlayerLives();

        this._manageUfoShips();

        for (let currentObjIdx: number = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject: GameObject = this._gameObjects[currentObjIdx];

            // In case of bullets, spaceship and ufoships, check for collisions with other objects
            if (gameObject instanceof Bullet || gameObject instanceof Spaceship || gameObject instanceof UfoShip)
                this._checkCollisions(gameObject);

            // Check if the object has been destroyed (after collision detection code)
            if (this._spaceShip.destroyed) {
                this._playerSpaceshipHit(currentObjIdx);
            }

            if (gameObject.destroyed) {
                this._gameObjects.splice(currentObjIdx, 1);

                if (gameObject instanceof UfoShip)
                    this._removeUfoShip(gameObject);
            }
            else {
                // Update object position and check if outside the game area
                // If so bring it back inside!
                gameObject.updatePosition();

                if (gameObject.position.x > this._gameAreaWidth)
                    gameObject.position.x = 0;
                else if (gameObject.position.x < 0)
                    gameObject.position.x = this._gameAreaWidth;
                
                if (gameObject.position.y > this._gameAreaHeight)
                    gameObject.position.y = 0;
                else if (gameObject.position.y < 0)
                    gameObject.position.y = this._gameAreaHeight;            

                // Draw the object
                gameObject.draw();
            }

            this._drawPlayerScore();

            if (this._gameObjects.filter(t => t instanceof Asteroid).length === 0) {
                // No more asteroids, level completed
                this._switchToPlayMode();
                return;
            }
        }
    }

    private _gameModeGameOver(): void {
        this._clearGameArea();

        if (this._checkKeyDown("Enter")) {
            this._resetKeyDown("Enter");
            this._switchToStartScreenMode();
            return;
        }

        for (let currentObjIdx: number = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject: GameObject = this._gameObjects[currentObjIdx];
            
            gameObject.updatePosition();
            
            gameObject.draw();
            
            if (gameObject.position.x > this._gameAreaWidth)
                gameObject.position.x = 0;
            else if (gameObject.position.x < 0)
                gameObject.position.x = this._gameAreaWidth;
            
            if (gameObject.position.y > this._gameAreaHeight)
                gameObject.position.y = 0;
            else if (gameObject.position.y < 0)
                gameObject.position.y = this._gameAreaHeight;
        }

        this._context.font = "100px Courier New";
        this._context.fillStyle = "white";
        this._context.textAlign = "center";
        this._context.fillText("Game Over", this._canvas.width / 2, this._canvas.height / 2);
        this._context.font = "25px Courier New";
        this._context.fillText("Press Enter to continue", this._canvas.width / 2, (this._canvas.height / 2) + 60);
    }

    private _updateGameArea(): void {
        if (this._gameMode === GameMode.startScreen)
            this._gameModeStartScreen();
        else if (this._gameMode === GameMode.gameOverScreen)
            this._gameModeGameOver();
        else
            this._gameModePlay();
    }

    private _onKeyDown(e: KeyboardEvent): void {
        this._keyDown.set(e.key, true);
        this._keyUp.set(e.key, false);
        e.stopImmediatePropagation();
        e.preventDefault();
    }

    private _onKeyUp(e: KeyboardEvent): void {
        this._keyDown.set(e.key, false);
        this._keyUp.set(e.key, true);
        e.stopImmediatePropagation();
        e.preventDefault();
    }

    private _resizeWindow(e: Event): void {
        this._gameAreaWidth = this._canvas.width;
        this._gameAreaHeight = this._canvas.height;
    }

    public init(): void {
        this._body = <HTMLBodyElement>document.getElementsByTagName("body")[0];
        this._canvas = <HTMLCanvasElement>document.getElementById("GameArea");
        this._context = this._canvas.getContext("2d");

        this._body.addEventListener("keydown", (e: KeyboardEvent) => this._onKeyDown(e));
        this._body.addEventListener("keyup", (e: KeyboardEvent) => this._onKeyUp(e));
        window.addEventListener("resize", (e: Event) => this._resizeWindow(e));
        this._resizeWindow(null);

        this._axisOrigin = new Point(0, this._gameAreaHeight);
        this._drawHelper = new DrawHelper(this._context, this._axisOrigin);

        this._switchToStartScreenMode();

        setInterval(() => { this._updateGameArea(); }, AppParameters.gameAreaUpdateDelayMS);        
    }

    public _switchToStartScreenMode(): void {
        this._gameMode = GameMode.startScreen;

        this._playerLives = AppParameters.maxPlayerLives;
        this._playerScore = 0;
        this._gameObjects = [];
        this._ufoShips = [];

        for (let i: number = 0; i < AppParameters.maxAsteroids * 3; i++) {
            let asteroid: Asteroid = new Asteroid(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight)));
            asteroid.init();
            this._gameObjects.push(asteroid);
        }
    }

    public _switchToPlayMode(): void {
        this._gameMode = GameMode.play;

        this._gameObjects = [];
        this._ufoShips = [];

        this._spaceShip = new Spaceship(this._drawHelper, new Point(this._gameAreaWidth / 2, this._gameAreaHeight / 2));
        this._gameObjects.push(this._spaceShip); 

        for (let i: number = 0; i < AppParameters.maxAsteroids; i++) {
            this._gameObjects.push(new Asteroid(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight))));       
        }
                
        for (let g of this._gameObjects) {
            g.init();
        }
    }

    public _switchToGameOverScreenMode(): void {
        this._gameMode = GameMode.gameOverScreen;
    }
}

let GI: Game = new Game();
GI.init();

let GameArea: HTMLCanvasElement = <HTMLCanvasElement>document.getElementById("GameArea");
